@@ -1,4 +1,5 @@
 #line 1
+#line 1
 package Module::Install::PRIVATE::Net::SSLeay;
 
 use strict;
@@ -118,6 +119,23 @@ sub ssleay_is_rsaref {
     return $ENV{OPENSSL_RSAREF};
 }
 
+my $other_try = 0;
+my @nopath;
+sub check_no_path {            # On OS/2 it would be typically on default paths
+    my $p;
+    if (not($other_try++) and $] >= 5.008001) {
+       require ExtUtils::Liblist;              # Buggy before this
+       my ($list) = ExtUtils::Liblist->ext("-lssl");
+       return unless $list =~ /-lssl\b/;
+        for $p (split /\Q$Config{path_sep}/, $ENV{PATH}) {
+           @nopath = ("$p/openssl$Config{_exe}",       # exe name
+                      '.')             # dummy lib path
+               if -x "$p/openssl$Config{_exe}"
+       }
+    }
+    @nopath;
+}
+
 sub find_openssl_prefix {
     my ($self, $dir) = @_;
 
@@ -135,6 +153,8 @@ sub find_openssl_prefix {
             '/apps/openssl/std/bin/openssl'  => '/apps/openssl/std',
 	    '/usr/sfw/bin/openssl'           => '/usr/sfw', # Open Solaris
             'C:\OpenSSL\bin\openssl.exe'     => 'C:\OpenSSL',
+            $Config{prefix} . '\bin\openssl.exe'      => $Config{prefix},           # strawberry perl
+            $Config{prefix} . '\..\c\bin\openssl.exe' => $Config{prefix} . '\..\c', # strawberry perl
     );
 
     while (my ($k, $v) = each %guesses) {
@@ -142,6 +162,8 @@ sub find_openssl_prefix {
             return $v;
         }
     }
+    (undef, $dir) = $self->check_no_path
+       and return $dir;
 
     return;
 }
@@ -156,6 +178,9 @@ sub find_openssl_exec {
             return $path;
         }
     }
+    ($prefix) = $self->check_no_path
+       and return $prefix;
+    return;
 }
 
 sub check_openssl_version {
